#pragma once

#include "storm.h"
#include "wc3/array.hpp"
#include "wc3/linkedlist.hpp"
#include <cstring>
#include <utility>

template<typename T, typename K>
class TSHashObject {
public:
	using key_type = typename K::key_type;

	uint32_t hash;
	LinkedNode<T> hash_node1;
	LinkedNode<T> hash_node2;
	key_type hash_key;
};

class HASHKEY_DUMMY {
public:
	using key_type = uint32_t;

	static bool is_keys_equal(const key_type& a, const key_type& b) {
		throw "Before using a hash table, configure the key type.";
	}

	static uint32_t hash(const key_type& key) {
		throw "Before using a hash table, configure the key type.";
	}
};

template<typename T = void*, typename K = HASHKEY_DUMMY>
class TSHashTable {
public:
	using key_type = typename K::key_type;

	class Iterator {
	public:
		typedef std::bidirectional_iterator_tag iterator_category;
		typedef std::pair<key_type, T*>          value_type;
		typedef ptrdiff_t                       difference_type;
		typedef value_type*                     pointer;
		typedef value_type&                     reference;
	private:
		friend class TSHashTable;

		const typename TSLinkedList<T>::Iterator item;

		Iterator(const typename TSLinkedList<T>::iterator item):item(item) {}
	public:
		bool operator == (const Iterator& other) const {
			return item == other.item;
		}

		bool operator != (const Iterator& other) const {
			return !(*this == other);
		}

		const value_type operator * () const {
			return {item->hash_key, item};
		}

		Iterator& operator ++ () {
			++item;
			return *this;
		}

		Iterator operator ++ (int) {
			Iterator temp = *this;
			++(*this);
			return temp;
		}

		Iterator& operator -- () {
			--item;
			return *this;
		}

		Iterator operator -- (int) {
			Iterator temp = *this;
			--(*this);
			return temp;
		}
	};

	struct VMT {
		void (__thiscall *erase_item)(TSHashTable* self, T* item);
		T* (__thiscall *create_item)(TSHashTable* self, TSLinkedList<T>* list, size_t extra_size, int extra_flags);
		void (__thiscall *operator_delete)(TSHashTable* self, bool free_memory);
		void (__thiscall *clear)(TSHashTable* self);
	};

	void* vtable;
	TSLinkedList<T> items;
	int counter;
	TSGrowableArray<TSLinkedList<T>> pages;
	uint32_t index_mask;

	auto find(const key_type& key, uint32_t hash) {
		if (index_mask != -1) {
			TSLinkedList<T>& page = pages[hash & index_mask];
			for (auto item = page.begin(); item != page.end(); ++item) {
				if (item->hash == hash && K::is_keys_equal(item->hash_key, key)) {
					return Iterator(item);
				}
			}
		}

		return end();
	}

	auto begin() {
		return Iterator(items.begin());
	}

	auto end() {
		return Iterator(items.end());
	}

	auto find(const key_type& key) {
		return find(key, K::hash(key));
	}
};

uint32_t hash32(uint32_t value);

class HASHKEY_NONE {
public:
	using key_type = uint32_t;

	static bool is_keys_equal(const key_type& a, const key_type& b) {
		return a == b;
	}

	static uint32_t hash(const key_type& key) {
		return key;
	}
};

template<typename K, bool case_sensetive>
class HASHKEY_STR_BASE {
public:
	using key_type = K;

	static bool is_keys_equal(const key_type& a, const key_type& b) {
		return (a == b) || case_sensetive ? strcmp(a, b) : stricmp(a, b);
	}

	static uint32_t hash(const key_type& key) {
		return SStringHash(key);
	}
};

using HASHKEY_STR = HASHKEY_STR_BASE<char*, true>;
using HASHKEY_STRI = HASHKEY_STR_BASE<char*, false>;
using HASHKEY_CONSTSTR = HASHKEY_STR_BASE<const char*, true>;
using HASHKEY_CONSTSTRI = HASHKEY_STR_BASE<const char*, false>;

class HASHKEY_4CC {
public:
	using key_type = uint32_t;

	static bool is_keys_equal(const key_type& a, const key_type& b) {
		return a == b;
	}

	static uint32_t hash(const key_type& key) {
		return hash32(key);
	}
};

class HASHKEY_8CC {
public:
	using key_type = uint64_t;

	static bool is_keys_equal(const key_type& a, const key_type& b) {
		return a == b;
	}

	static uint32_t hash(const key_type& key) {
		throw "HASHKEY_8CC::hash not yet implemented.";
	}
};
