#include "wc3/game_base.hpp"

#include <algorithm>
#include <iterator>
#include <libloaderapi.h>
#include <cstdio>
#include <errhandlingapi.h>
#include <array>
#include <optional>
#include <winver.h>

typedef std::array<WORD, 4> FileVersion;

static constexpr std::pair<FileVersion, GameVersion> game_version_table[] {
	{{1, 26, 0, 6401}, GameVersion::v1_26a},
	{{1, 27, 0, 52240}, GameVersion::v1_27a},
};

static std::optional<GameVersion> game_version;

const char* get_game_version_name(GameVersion version) {
	switch (version) {
		case GameVersion::v1_26a: return "1.26a";
		case GameVersion::v1_27a: return "1.27a";
		default: return "unknown";
	}
}

uintptr_t get_offset(const char* name, GameVersion version, const Offsets& offsets) {
	auto it = offsets.find(version);
	if (it != offsets.end()) {
		return it->second;
	}
	else {
		fprintf(
			stderr,
			"Warning! Offset «%s» was not found for %s version of the game.\n",
			name,
			get_game_version_name(version)
		);

		return 0;
	}
}

uintptr_t get_offset(const char* name, const Offsets& offsets) {
	return get_offset(name, get_game_version(), offsets);
}

bool get_file_version(const char* file_path, FileVersion& version) {
	DWORD info_size = GetFileVersionInfoSize(file_path, nullptr);
	if (!info_size) {
		fprintf(
			stderr,
			"GetFileVersionInfoSize for «%s» failed. Error = %d.\n",
			file_path,
			GetLastError()
		);
		return false;
	}

	unsigned char info_buffer[info_size];

	if (!GetFileVersionInfo(
		file_path,
		0,
		info_size,
		&info_buffer
	)) {
		fprintf(
			stderr,
			"GetFileVersionInfo for «%s» failed. Error = %d.\n",
			file_path,
			GetLastError()
		);
		return false;
	}

	VS_FIXEDFILEINFO* fixed_info;
	UINT fixed_info_size;

	if (!VerQueryValue(
		info_buffer,
		"\\",
		(LPVOID*)&fixed_info,
		&fixed_info_size

	)) {
		fprintf(
			stderr,
			"VerQueryValue for «%s» failed. Error = %d.\n",
			file_path,
			GetLastError()
		);
		return false;
	};

	version[0] = HIWORD(fixed_info->dwFileVersionMS);
	version[1] = LOWORD(fixed_info->dwFileVersionMS);
	version[2] = HIWORD(fixed_info->dwFileVersionLS);
	version[3] = LOWORD(fixed_info->dwFileVersionLS);

	return true;
}

GameVersion get_game_version() {
	if (!game_version.has_value()) {
		FileVersion version;
		if (get_file_version("game.dll", version)) {
			auto it = std::find_if(
				std::begin(game_version_table),
				std::end(game_version_table),
				[&] (auto& x) { return x.first == version; }
			);
			if (it != std::end(game_version_table)) {
				game_version = it->second;
			}
			else {
				fprintf(
					stderr,
					"Unknown game version: %d.%d.%d.%d.\n",
					version[0],
					version[1],
					version[2],
					version[3]
				);
				game_version = GameVersion::unknown;

				printf("Known versions:\n");
				for (auto& [fv, gv] : game_version_table) {
					printf(
						"%d.%d.%d.%d %s\n",
						fv[0], fv[1], fv[2], fv[3],
						get_game_version_name(gv)
					);
				}
			}
		}
		else {
			fprintf(
				stderr,
				"Failed to get version of «game.dll». Error = %d.",
				GetLastError()
			);
			game_version = GameVersion::unknown;
		}
	}

	return game_version.value();
}

uintptr_t get_game_base() {
	static uintptr_t game_base;

	if (!game_base) {
		game_base = (uintptr_t)GetModuleHandle("game.dll");
	}

	return game_base;
}
