#include "wc3/net_data.hpp"

#include <algorithm>
#include <cstring>
#include <tuple>
#include "storm.h"
#include "wc3/net_provider.hpp"

const size_t TURN_STORE_RESERVED_BYTES = 8;
const size_t TURN_STORE_LIMIT = 1024;

void CTurnStore::flush() {
	if (size > TURN_STORE_RESERVED_BYTES) {
		void* commands = (char*)data + TURN_STORE_RESERVED_BYTES;
		size_t commands_size = size - TURN_STORE_RESERVED_BYTES;
		NetProvider::instance()->net_client->send(commands, commands_size);
		clear();
	}
}

CNetData::CNetEvent::CNetEvent(void* buffer, size_t size, size_t capacity, uint8_t id, uint8_t unk) {
	this->buffer = buffer;
	this->size = size;
	this->capacity = capacity;
	this->id = id;
	this->unk = unk;
}

void CNetData::CNetEventQueue::add(CNetData::CNetEvent* event) {
	auto target = events.end();

	if (event->unk == 2) {
		auto it = std::find_if(
			events.rbegin(),
			events.rend(),
			[event] (CNetEvent& x) {
				return x.unk == event->unk;
			}
		);
		if (it != events.rend()) {
			target = (++it).base();
		}
	}

	events.insert(*event, target);
}

std::tuple<void*, size_t> get_net_memory_from_cache(size_t required_size) {
	void* buffer;
	size_t size;

	NetProvider* provider = NetProvider::instance();

	if (!provider || !provider->recycler.get_data(buffer, size)) {
		buffer = SMemAlloc(required_size, __FILE_NAME__, __LINE__, 0);
		size = required_size;
	}
	else if (size < required_size) {
		buffer = SMemRealloc(buffer, required_size, __FILE_NAME__, __LINE__, 0);
		size = required_size;
	}

	return { buffer, size };
}

struct IncomingActionsEventData {
	uint16_t commands_size_to_log;
	uint8_t player_id;
	uint16_t commands_size;
	uint8_t commands[];
};

void CNetData::_send_turn_sp() {
	CNetSession* session = &sessions[(int)current_session_type];

	if (!session->suspended_players.size()
		&& !net_event_queue.net_event_counters[current_session_type == CNetSession::Type::Game][(int)NetEventType::HZ_24h]
	) {
		void* commands = (char*)turn_store.data + turn_reserved_bytes;
		size_t commands_size = turn_store.size - turn_reserved_bytes;

		if (commands_size) {
			size_t required_size = sizeof(IncomingActionsEventData) + commands_size;
			auto [memory, capacity] = get_net_memory_from_cache(required_size);

			IncomingActionsEventData* event_data = reinterpret_cast<IncomingActionsEventData*>(memory);
			event_data->commands_size_to_log = 0;
			event_data->player_id = session->players.get_start_loc_(session->local_player_id);
			event_data->commands_size = commands_size;
			memcpy(event_data->commands, commands, commands_size);

			CNetEvent* event = net_event_recycler.get_item(
				event_data,
				required_size,
				capacity,
				(uint8_t)NetEventType::INCOMING_ACTIONS,
				2
			);

			net_event_queue.add(event);
			enqueued_commands_size += event_data->commands_size_to_log;
			turn_store.clear();
		}
	}
}

void CNetData::_send_turn() {
	switch (net_provider_id) {
		case 'LOOP':
		case 'NONE':
			_send_turn_sp();
			break;
		default:
			turn_store.flush();
			break;
	}
}

bool CNetData::send_command(CDataStore& store, CNetSession::Type session_type) {
	if (current_session_type == session_type
		&& sessions[(int)session_type].state >= CNetSession::GameState::LOADING
	) {
		if ((turn_store.size - turn_reserved_bytes + store.size) >= TURN_STORE_LIMIT) {
			_send_turn();
		}

		if ((turn_store.size - turn_reserved_bytes + store.size) < TURN_STORE_LIMIT) {
			turn_store.position = turn_store.size;
			turn_store.write(store.data, store.size);
			return true;
		}
		else {
			fprintf(stderr, "%s(%d): rejected %u bytes\n", __FUNCTION__, __LINE__, store.size);
		}
	}

	return false;
}
